/*
 * LSU -- local superuser
 *
 * Author:
 *	Matt Bishop
 *	Research Institute for Advanced Computer Science
 *	NASA Ames Research Center
 *	Moffett Field, CA  94035
 *
 *	mab@riacs.edu			ARPA
 *	...!decvax!decwrl!riacs!mab	UUCP
 *	...!ihnp4!ames!riacs!mab	UUCP
 *
 * Copyright (c) 1986.
 */
#include "lsu.h"

/*
 * global variables
 */
char *progname = NULL;		/* program name */
struct perms *progis;		/* what the program runs as */
struct perms runas[] = {	/* legal program names -- MUST be one of these */
	{ "SU", "lsu", },		/* local super user */
	{ "su", "su",  },		/* super user */
	{ "nu", "nsu", },		/* new super user */
	{ "cu", "csu", },		/* check access file syntax */
	{ "??", NULL,  },		/* end of list marker */
};
FILE *logfp = NULL;		/* log file pointer */
char *date;			/* date of run */
char *tty;			/* terminal name or background */
char *towho = NULL;		/* command line argument */
struct passwd curpw;		/* information about who's running the program */
struct passwd newpw;		/* information about who you're lsu'ing to */
char *shell = LSUSHELL;		/* shell to be exec'ed */
int stamp = -1;			/* PID stamp of this process (for logging) */
int dash = 0;			/* 1 if a login shell to be used */
struct sv shvar[] = LSUVARS;	/* list of shell variables to change */
unsigned int success = F_NONE;	/* F_NONE if l/n/csu is to succeed */
#ifdef lint
				/*
				 * these are here to shut lint up; they are
				 * used for system errors
				 */
int errno;			/* system error number */
int sys_nerr;			/* number of elements in sys_errlist[] */
char *sys_errlist[1];		/* list of system error messages */
#else
				/*
				 * these are used to keep track
				 * of how and when the binaries
				 * were made
				 */
char *lsu_system = SYSTEM;	/* what operating system this was made for */
char *lsu_version = VERSION;	/* what version this is */
#endif

main(argc, argv, envp)
int argc;			/* number of arguments */
char **argv;			/* argument list */
char **envp;			/* environment list */
{
	char buf[BUFSIZ];		/* buffer for error and other messages */

	/*
	 * set up the password fields (used for error messages
	 * if no password yet demanded), open the error log,
	 * set up the program's name, and initialize the
	 * structures this goodie needs
	 */
	init();
	setname(argv[0]);
	openlog();

	/*
	 * if checking syntax only, do so and stop
	 */
	if (progname == CSU)
		(void) exit(permsyntax(++argv));

	/*
	 * process any arguments
	 */
	if (argc > 1 && strcmp(argv[1], "-") == 0){
	       	dash = 1;
		argc--;
		argv++;
	}
	if (argc > 1){
		if (strncmp(argv[1], "--", 2) != 0)
			towho = strsave(argv[1]);
		argc--;
		argv++;
	}

	/*
	 * figure out who to become
	 */
	if (towho != NULL)
		(void) strcpy(buf, towho);
	else if (progname == SU || progname == NSU)
		(void) strcpy(buf, LSUUSER);
	else
		buf[0] = '\0';

	/*
	 * see if the user has permission to change to
	 * whoever he/she wants
	 */
	perms(buf);

	/*
	 * at this point buf[] contains the new user
	 * gather new user data and check the password
	 */
	getnewuser(buf);
	chkpasswd();

	/*
	 * get the right shell
	 * if that fails, assume the default shell
	 * also reset the argument list
	 */
	getshell(argv, envp);

	/*
	 * reset the path environment variable if need be
	 */
	if (newpw.pw_uid == 0)
		chkpath(&envp);

	/*
	 * reset the UID and GID of this process and log it
	 */
	if (setgid(newpw.pw_gid) < 0)
		err(F_COND|F_SYS, buf);
	if (setuid(newpw.pw_uid) < 0)
		err(F_COND|F_SYS, buf);

	/*
	 * check for success or failure here
	 */
	switch(success){
	case F_COND:
		if ((progname == NSU || progname == SU) &&
					(newpw.pw_uid == SUPERUID))
			break;
		/* FALLS THROUGH */
	case F_FAIL:
		logmessage(L_NO, "permission denied\n");
		fprintf(stderr, "%s: permission denied\n", progname);
		exit(1);
	default:
		break;
	}
	(void) sprintf(buf, "became %s (UID %d, GID %d)\n",
				newpw.pw_name, newpw.pw_uid, newpw.pw_gid);
	logmessage(L_YES, buf);
	if (logfp != NULL)
		(void) fclose(logfp);

	/*
	 * now overlay the shell
	 */
	execve(shell, argv, envp);

	/*
	 * if you get here the execve failed, so ...
	 */
	openlog();
	err(F_COND|F_SYS, "execve");
	/* NOTREACHED */
}

/*
 * initialize -- get the user information and the date
 *		 and open the log file for appending
 */
init()
{
	long clock;		/* internal representation of current time */
	char *t;		/* used to locate terminal */
	char buf[BUFSIZ];	/* error message buffer */
	struct passwd *pw;	/* password file entry */
	struct tm *tick;	/* local time */

	/*
	 * get the stamp (for errors and logging)
	 */
	stamp = getpid();

	/*
	 * get the date
	 */
	clock = time((long *) 0);
	tick = localtime(&clock);
	(void) sprintf(buf, "%02d/%02d %02d:%02d",
		tick->tm_mon+1, tick->tm_mday, tick->tm_hour + 1, tick->tm_min);
	date = strsave(buf);

	/*
	 * get the terminal number
	 * if none available, it's being run from background
	 */
	if ((t = ttyname(0)) == NULL && (t = ttyname(1)) == NULL &&
						(t = ttyname(2)) == NULL)
		t = "background";
	if (strncmp(t, "/dev/", strlen("/dev/")) == 0)
		t += strlen("/dev/");
	tty = strsave(t);

	/*
	 * get the password file information for this user
	 * if none, log the error using the current UID and die
	 */
	curpw.pw_name = NULL;
	curpw.pw_uid = getuid();
	curpw.pw_gid = getgid();
	if ((pw = getpwuid(curpw.pw_uid)) == NULL){
		(void) sprintf(buf, "no user with UID %d\n", curpw.pw_uid);
		err(F_COND, buf);
	}
	curpw.pw_name = strsave(pw->pw_name);
	curpw.pw_passwd = strsave(pw->pw_passwd);
	curpw.pw_dir = strsave(pw->pw_dir);
	curpw.pw_shell = strsave(pw->pw_shell);
}

/*
 * get shell -- set up argv0 and shell
 */
getshell(argv, envp)
char **argv;			/* argument list */
char **envp;			/* environment list */
{
        register int i, j;      /* counter in for loops */
	char *p;		/* used to set up arg 0 to shell */
	char *argv0;            /* argument 0 */
	char tmpbuf[BUFSIZ];	/* temporary buffer */

	/*
	 * set up the shell
	 */
	if (newpw.pw_shell[0] != '\0')
		shell = strsave(newpw.pw_shell);
	else{
		(void) free(newpw.pw_shell);
		newpw.pw_shell = strsave(LSUSHELL);
		shell = strsave(LSUSHELL);
	}

	/*
	 * get the last part of the program name
	 */
	argv0 = strsave(progname);
	if ((p = rindex(argv0, '/')) == NULL){
		(void) sprintf(tmpbuf, "/%s", progname);
		p = tmpbuf;
	}

	/*
	 * if this is to be a login shell, arg 0 goes to '-'
	 */
	if (dash)
		*p = '-';
	else
		p++;

	/*
	 * now stuff this in argument 0
	 */
	argv[0] = strsave(p);

	/*
	 * set up the environment
	 */
	for(j = 0; envp[j] != NULL; j++){
		for(i = 0; shvar[i].splate != NULL; i++){
			if (strncmp(envp[j], shvar[i].splate, shvar[i].len) == 0 &&
							envdoit(shvar[i].vused)){
				(void) sprintf(tmpbuf, shvar[i].splate, *shvar[i].cval);
				envp[j] = strsave(tmpbuf);
			}
		}
	}
}

/*
 * determine if a shell variable should be replaced
 */
int envdoit(when)
unsigned int when;
{
	/*
	 * if always done, succeed
	 */
	if (bitset(when, V_ALWAYS))
		return(1);
	/*
	 * if never done, fail
	 */
	if (bitset(when, V_NONE))
		return(0);
	/*
	 * if done only for a login shell and this isn't one, fail
	 */
	if (bitset(when, V_ANDLOGIN) && !dash)
		return(0);
	/*
	 * if done for a login shell or anything else,
	 * and this is a login shell, succeed
	 */
	if (bitset(when, V_ORLOGIN) && dash)
		return(1);
	/*
	 * on a per-program basis
	 */
	if (bitset(when, V_CSU) && progname == CSU)
		return(1);
	if (bitset(when, V_LSU) && progname == LSU)
		return(1);
	if (bitset(when, V_NSU) && progname == NSU)
		return(1);
	if (bitset(when, V_SU) && progname == SU)
		return(1);
	/*
	 * failure
	 */
	return(0);
}

/*
 * reset the PATH variable to a default, safe one
 */
chkpath(envp)
char ***envp;
{
	register int i, j;	/* counter in a for loop */
	union {			/* used to allocate space for new environment */
		char **cpp;		/* char ** space */
		char *cp;		/* char * space */
	} u;			/* the union */

	/*
	 * see if there's a PATH environment variable
	 */
	for(i = 0; (*envp)[i] != NULL; i++)
		if (strncmp((*envp)[i], "PATH=", 5) == 0){
			/*
			 * there is; put in the new path and return
			 */
			(*envp)[i] = strsave(LSUPATH);
			return;
		}

	/*
	 * allocate space for a new environment
	 */
	if ((u.cp = malloc((unsigned) (sizeof(char **) * (i+2)))) == NULL)
		err(F_SYS|F_COND, "environment");

	/*
	 * copy the environment over
	 */
	for(j = 0, i = 0; (*envp)[i] != NULL; i++)
		u.cpp[j++] = (*envp)[i];

	/*
	 * insert the new path
	 */
	u.cpp[j++] = strsave(LSUPATH);

	/*
	 * end the new environment
	 * and save it where the old one is
	 */
	u.cpp[j] = NULL;
	*envp = u.cpp;
}

