/*
 * LSU -- local superuser header file
 *
 * Author:
 *	Matt Bishop
 *	Research Institute for Advanced Computer Science
 *	NASA Ames Research Center
 *	Moffett Field, CA  94035
 *
 *	mab@riacs.edu			ARPA
 *	...!decvax!decwrl!riacs!mab	UUCP
 *	...!ihnp4!ames!riacs!mab	UUCP
 *
 * Copyright (c) 1986.
 */
#include "lsu.h"

/*
 * line information; used to print errors
 */
extern int linect;			/* line number (see "perm.c") */

/*
 * this is the driver that checks the access file syntax
 * it calls a checker for each file in the argument list
 */
permsyntax(argv)
char **argv;			/* argument list (argv[1]->...) */
{
	register int i;			/* counter in a for loop */
	register int res = 0;		/* return value */
	register FILE *fp;		/* file to be checked */
	char buf[BUFSIZ];		/* buffer for messages */

	/*
	 * if not root, usual "permission denied" stuff
	 */
	if (curpw.pw_uid != 0){
		logmessage(L_CHK, "permission denied (not root)\n");
		err(L_CHK|F_FAIL, "permission denied\n");
		return(1);
	}

	/*
	 * no file named -- use standard input
	 */
	if (*argv == NULL)
		return(psyn(stdin));
	/*
	 * walk the list
	 */
	for(i = 0; argv[i] != NULL; i++){
		/*
		 * log the check
		 */
		(void) sprintf(buf, "checking syntax of %s\n", argv[i]);
		logmessage(L_CHK, buf);
		/*
		 * open the file; on failure, give error message
		 */
		if ((fp = fopen(argv[i], "r")) == NULL){
			err(L_INFO|F_SYS, argv[i]);
			res = 1;
		}
		else{
			/*
			 * print the name of the file being scanned
			 * and scan it
			 */
			fprintf(stderr, "%s:\n", argv[i]);
			psyn(fp);
		}
	}
	/*
	 * return the result of the scan(s)
	 */
	return(res);
}

/*
 * check the syntax of one file
 */
psyn(fp)
FILE *fp;			/* file pointer */
{
	char buf[BUFSIZ];		/* buffer for line */
	char ebuf[BUFSIZ];		/* buffer for error messages */
	register char *b, *f;		/* used to get fields */
	char c;				/* character ending a name */
	int exstat = 0;			/* exit status */

	/*
	 * process the file a line at a time
	 */
	for(linect = 1; fgets(buf, BUFSIZ, fp) != NULL; linect++){
		/*
		 * comment lines are correct
		 */
		if (buf[0] == LSUCOMM)
			continue;
		/*
		 * eat leading spaces; if nothing else,
		 * the line is empty
		 */
		for(b = buf; isspace(*b); b++);
		if (*b == '\0'){
			(void) sprintf(ebuf, "(%2d): empty line\n", linect);
			err(L_INFO|F_NONE, ebuf);
			exstat = 1;
			continue;
		}
		/*
		 * begin user field; f points to user name,
		 * b will point to just beyond end of this field
		 */
		f = b;
		while(*b && !isspace(*b))
			b++;
		if (isspace(*b))
			*b++ = '\0';
		/*
		 * if no such user, warn -- may be problem
		 */
		if (getpwnam(f) == PNULL){
			(void) sprintf(ebuf, "(%2d): %s -- no such user\n",
								linect, f);
			err(L_INFO|F_NONE, ebuf);
			exstat = 1;
		}
		/*
		 * skip to next field; if nothing else,
		 * there are three fields missing
		 */
		while(*b != '\t')
			b++;
		while(isspace(*b))
			b++;
		if (*b == '\0'){
			(void) sprintf(ebuf,
				"(%2d): missing newuser, tty, time fields\n",
								linect);
			err(L_INFO|F_NONE, ebuf);
			exstat = 1;
			continue;
		}
		/*
		 * begin newuser user field; f points to a newuser name,
		 * b will point to just beyond end of that name
		 */
		f = b;
		do{ 
			/*
			 * skip over the name; if it ends in comma,
			 * another follows
			 */
			while(*b && *b != ',' && !isspace(*b))
				b++;
			c = *b;
			*b++ = '\0';
			/*
			 * is it a legal user
			 */
			if (!isany(f) && getpwnam(f) == PNULL){
				(void) sprintf(ebuf, "(%2d): %s -- no such user\n",
								linect, f);
				exstat = 1;
				err(L_INFO|F_NONE, ebuf);
			}
			/*
			 * now look at beginning of next one
			 */
			f = b;
		} while(c == ',');
		/*
		 * skip to next field; if nothing else,
		 * there are two fields missing
		 */
		if (c != '\t')
			while(*b && *b != '\t')
				b++;
		while(isspace(*b))
			b++;
		if (*b == '\0'){
			(void) sprintf(ebuf, "(%2d): missing tty, time fields\n",
								linect);
			err(L_INFO|F_NONE, ebuf);
			exstat = 1;
			continue;
		}
		/*
		 * now check the terminal specification syntax
		 */
		f = b;
		(void) isintty(f);
		/*
		 * skip to next field; if nothing else,
		 * there are two fields missing
		 */
		while(*b && *b != '\t')
			b++;
		while(isspace(*b))
			b++;
		if (*b == '\0'){
			(void) sprintf(ebuf, "(%2d): missing time field\n",
								linect);
			err(L_INFO|F_NONE, ebuf);
			continue;
		}
		/*
		 * now check the time specification format
		 */
		f = b;
		(void) isintime(f);
	}
	return(exstat);
}

/*
 * this returns 1 if the word f is "ANY" in any case
 */
int isany(f)
char *f;
{
	/*
	 * you maybe expected a table lookup?
	 */
	return (f[0] && tolower(f[0]) == 'a'
		&& f[1] && tolower(f[1]) == 'n'
			&& f[2] && tolower(f[2]) == 'y'
				&& f[3] == '\0');
}

