%{
/*
 * this file contains the parser for the time of day
 * the function "isintime(t)" returns 1 if the current time
 * is within the time given by the string t
 * grammar:
 *	day_of_year day_of_week time_of_day
 * where:
 *	day_of_year == month number ',' number		(ie, Aug 29, 1986)
 *		       month number			(ie, Aug 29)
 *		       month ',' number			(ie, Aug, 1986)
 *		       month				(ie, Aug)
 *		       number '/' number '/' number	(ie, 8/29/86)
 *		       number '/' number		(ie, 8/29)
 *	day_of_week == special_day
 *			Any of "Sunday", "Monday", "Tuesday",
 *			"Wednesday", "Thursday", "Friday",
 *			or "Saturday"
 *	time_of_day == number ':' number ':' number mer (ie, 8:12:56 pm)
 *		       number ':' number ':' number	(ie, 20:12:56)
 *		       number ':' number meridian	(ie, 8:12 pm)
 *		       number ':' number 		(ie, 20:12)
 *		       number meridian			(ie, 8 pm)
 *		       number  				(ie, 20)
 *		       special_time
 *			Any of "noon", "midnight"
 * "meridian" is "am" or "pm".  Enough of "special_day" and "special_time" words
 * to identify the word uniquely must be given.
 *
 * if you want to change the syntax, you can debug the new grammar by defining
 * "DEBUG" -- this allows the file to be compiled as a separate program, and
 * will display the result.
 *
 * Author:
 *	Matt Bishop
 *	Research Institute for Advanced Computer Science
 *	NASA Ames Research Center
 *	Moffett Field, CA  94035
 *
 *	mab@riacs.arpa			ARPA
 *	...!decvax!decwrl!riacs!mab	UUCP
 *	...!ihnp4!ames!riacs!mab	UUCP
 *
 * Copyright (c) 1986.
 *
 */
#include <ctype.h>
#include <stdio.h>


#ifdef DEBUG
#	define	F_COND	0	/* dummy value; not used here */
#	define	L_INFO	0	/* dummy value; not used here */
#	include "sysdep.h"
#else
#	include "lsu.h"
#endif

/*
 * useful macros
 */
				/* 1 if current time is between x and y */
#define intime(x,y)	(timecmp((x), &cur) != 1 && timecmp(&cur, (y)) != 1)
				/* if c is upper case, make it lower case */
#define lowcase(c)	(isupper((c))?tolower((c)):(c))
#define LK_AMBIG	-1	/* ambiguous result from table lookup */
#define LK_NOSUCH	-2	/* no such result from table lookup */

/*
 * times are represented as intervals; lo < hi
 */
typedef struct {
	struct tm lo;		/* lower bound of time */
	struct tm hi;		/* upper bound of time */
} TIME;
#define NULL_TIME	((TIME *) NULL)

%}

/*
 * for reasons due entirely to the way YACC processes input,
 * this has to go here since TIME is used as a type in the
 * header file to declare yylval
 */
%union {
	int ival;	/* as an integer */
	TIME vval;	/* as a time interval */
}

/*
 * tokens returned by the lexical analyzer yylex()
 */
%token <ival> AND		/* integer:  ampersand */
%token <ival> ANY		/* integer:  matches any date */
%token <ival> COMMA		/* integer:  comma */
%token <ival> COLON		/* integer:  colon */
%token <ival> DASH		/* integer:  hyphen */
%token <vval> DAY_OF_WEEK	/* interval: day of the week */
%token <ival> EOL		/* integer:  no more input */
%token <ival> LPAR		/* integer:  left parenthesis */
%token <vval> MERIDIAN		/* interval: am or pm */
%token <vval> MONTH		/* interval: month of the year */
%token <ival> NONE		/* integer:  matches no date */
%token <ival> NOT		/* integer:  exclamation point */
%token <ival> NUMBER		/* integer:  number token */
%token <ival> OR		/* integer:  vertical bar */
%token <ival> RPAR		/* integer:  right parenthesis */
%token <ival> SLASH		/* integer:  slash */
%token <vval> TIME_OF_DAY	/* interval: time of the day */
%token <ival> UNK		/* integer:  unknown */

/*
 * productions analyzed by the parser
 * in what follows, "interval" refers to the (given) time interval
 */
%type <ival> day_of_week_ok	/* integer:  1 if in day of week interval */
%type <vval> day_of_year	/* interval: bounds day of the year */
%type <ival> day_of_year_ok	/* integer:  1 if in day of year interval */
%type <ival> expr		/* integer:  1 if time so far in interval */
%type <ival> stat		/* integer:  expr with an EOL token tacked on */
%type <vval> time_of_day	/* interval: bounds time of the day */
%type <ival> time_of_day_ok	/* integer:  1 if in time of day interval */

/*
 * expression operators
 * these must be on the same line since they are of equal precedence
 */
%left OR AND
%right NOT

%{
#include "timeyd.h"

/*
 * time fields which we don't care about in comparisons are
 * set to -1 to warn the comparison function "timecmp" that
 * they are not to be considered; here's the prototype
 */
TIME nulltime = {
		{-1,-1,-1,-1,-1,-1,-1,-1,-1 }, 		/* lower bound */
		{-1,-1,-1,-1,-1,-1,-1,-1,-1 },		/* upper bound */
};

/*
 * lookup table -- all special words (keywords to compiler types) go here
 * the time limits go into the time table
 * (this should be one table, but some compilers -- like Amdahl's UTS
 * C compiler -- complain about too many initializers ...)
 */
struct looktbl {
	char *string;		/* the key word */
	int refnum;		/* number referring to keyword of this rettype */
	int rettype;		/* MONTH, MERIDIAN, TIME_OF_DAY, DAY_OF_WEEK */
} wordlist[] = {
	{ "any",	-1, ANY },		/*  0 */
	{ "am",		 0, MERIDIAN, },	/*  1 */
	{ "april",	 4, MONTH, },		/*  2 */
	{ "august",	 8, MONTH, },		/*  3 */
	{ "december",	12, MONTH, },		/*  4 */
	{ "february",	 2, MONTH, },		/*  5 */
	{ "friday",	 5, DAY_OF_WEEK, },	/*  6 */
	{ "january",	 1, MONTH, },		/*  7 */
	{ "july",	 7, MONTH, },		/*  8 */
	{ "june",	 6, MONTH, },		/*  9 */
	{ "march",	 3, MONTH, },		/* 10 */
	{ "may",	 5, MONTH, },		/* 11 */
	{ "midnight",	 0, TIME_OF_DAY, },	/* 12 */
	{ "monday",	 1, DAY_OF_WEEK, },	/* 13 */
	{ "none",	-1, NONE },		/* 14 */
	{ "noon",	12, TIME_OF_DAY, },	/* 15 */
	{ "november",	11, MONTH, },		/* 16 */
	{ "october",	10, MONTH, },		/* 17 */
	{ "pm",		12, MERIDIAN, },	/* 18 */
	{ "saturday",	 6, DAY_OF_WEEK, },	/* 19 */
	{ "september",	 9, MONTH, },		/* 20 */
	{ "sunday",	 0, DAY_OF_WEEK, },	/* 21 */
	{ "thursday",	 4, DAY_OF_WEEK, },	/* 22 */
	{ "tuesday",	 2, DAY_OF_WEEK, },	/* 23 */
	{ "wednesday",	 3, DAY_OF_WEEK, },	/* 24 */
	{ "weekdays",   -1, DAY_OF_WEEK, },	/* 25 */
	{ NULL,		 0, 0, },		/* -1 */
};
TIME timelist[] ={		/* lower and upper bounds of time period */
	{ {-1,-1,-1,-1,-1,-1,-1,-1,-1}, {-1,-1,-1,-1,-1,-1,-1,-1,-1} }, /*  0 */
	{ {-1,-1, 0,-1,-1,-1,-1,-1,-1}, {-1,-1, 0,-1,-1,-1,-1,-1,-1} }, /*  1 */
	{ { 0, 0, 0, 0, 3,-1,-1,-1,-1}, {60,60,24,30, 3,-1,-1,-1,-1} }, /*  2 */
	{ { 0, 0, 0, 0, 7,-1,-1,-1,-1}, {60,60,24,31, 7,-1,-1,-1,-1} }, /*  3 */
	{ { 0, 0, 0, 0,11,-1,-1,-1,-1}, {60,60,24,31,11,-1,-1,-1,-1} }, /*  4 */
	{ { 0, 0, 0, 0, 1,-1,-1,-1,-1}, {60,60,24,29, 1,-1,-1,-1,-1} }, /*  5 */
	{ { 0, 0, 0,-1,-1,-1, 5,-1,-1}, {60,60,24,-1,-1,-1, 5,-1,-1} }, /*  6 */
	{ { 0, 0, 0, 0, 0,-1,-1,-1,-1}, {60,60,24,31, 0,-1,-1,-1,-1} }, /*  7 */
	{ { 0, 0, 0, 0, 6,-1,-1,-1,-1}, {60,60,24,31, 6,-1,-1,-1,-1} }, /*  8 */
	{ { 0, 0, 0, 0, 5,-1,-1,-1,-1}, {60,60,24,30, 5,-1,-1,-1,-1} }, /*  9 */
	{ { 0, 0, 0, 0, 2,-1,-1,-1,-1}, {60,60,24,31, 2,-1,-1,-1,-1} }, /* 10 */
	{ { 0, 0, 0, 0, 4,-1,-1,-1,-1}, {60,60,24,31, 4,-1,-1,-1,-1} }, /* 11 */
	{ { 0, 0, 0,-1,-1,-1,-1,-1,-1}, { 0, 0, 0,-1,-1,-1,-1,-1,-1} }, /* 12 */
	{ { 0, 0, 0,-1,-1,-1, 1,-1,-1}, {60,60,24,-1,-1,-1, 1,-1,-1} }, /* 13 */
	{ {-1,-1,-1,-1,-1,-1,-1,-1,-1}, {-1,-1,-1,-1,-1,-1,-1,-1,-1} }, /* 14 */
	{ { 0, 0,12,-1,-1,-1,-1,-1,-1}, { 0, 0,12,-1,-1,-1,-1,-1,-1} }, /* 15 */
	{ { 0, 0, 0, 0,10,-1,-1,-1,-1}, {60,60,24,30,10,-1,-1,-1,-1} }, /* 16 */
	{ { 0, 0, 0, 0, 9,-1,-1,-1,-1}, {60,60,24,31, 9,-1,-1,-1,-1} }, /* 17 */
	{ {-1,-1,12,-1,-1,-1,-1,-1,-1}, {-1,-1,12,-1,-1,-1,-1,-1,-1} }, /* 18 */
	{ { 0, 0, 0,-1,-1,-1, 6,-1,-1}, {60,60,24,-1,-1,-1, 6,-1,-1} }, /* 19 */
	{ { 0, 0, 0, 0, 8,-1,-1,-1,-1}, {60,60,24,30, 8,-1,-1,-1,-1} }, /* 20 */
	{ { 0, 0, 0,-1,-1,-1, 0,-1,-1}, {60,60,24,-1,-1,-1, 0,-1,-1} }, /* 21 */
	{ { 0, 0, 0,-1,-1,-1, 4,-1,-1}, {60,60,24,-1,-1,-1, 4,-1,-1} }, /* 22 */
	{ { 0, 0, 0,-1,-1,-1, 2,-1,-1}, {60,60,24,-1,-1,-1, 2,-1,-1} }, /* 23 */
	{ { 0, 0, 0,-1,-1,-1, 3,-1,-1}, {60,60,24,-1,-1,-1, 3,-1,-1} }, /* 24 */
	{ { 0, 0, 0,-1,-1,-1, 1,-1,-1}, {60,60,24,-1,-1,-1, 5,-1,-1} }, /* 25 */
	{ {-1,-1,-1,-1,-1,-1,-1,-1,-1}, {-1,-1,-1,-1,-1,-1,-1,-1,-1} }, /* -1 */
};

/*
 * variables
 */
static int ateol = 0;	/* 1 when you hit the end of string */
static struct tm cur;	/* current time as structure */
static int timing;	/* 1 if time is okay, 0 if not */
static char *lptr;	/* used to walk input string */
static long clocktm;	/* internal time */

/*
 * library functions
 */
char *index();		/* find last character in a string */
struct tm *localtime();	/* convert internal time to a more useable structure */
long time();		/* return internal time */

%}

/*
 * start analysis at state stat
 */
%start stat

%%

stat		: expr EOL
			{ timing = $1; }
		;

expr		: LPAR expr RPAR
			{ $$ = $2; }
		| NOT expr
			{ $$ = !$2; }
		| expr OR expr
			{ $$ = $1 || $3; }
		| expr AND expr
			{ $$ = $1 && $3; }
		| day_of_year_ok day_of_week_ok time_of_day_ok
			{ $$ = $1 && $2 && $3; }
		| day_of_week_ok day_of_year_ok time_of_day_ok
			{ $$ = $1 && $2 && $3; }
		| day_of_year_ok time_of_day_ok
			{ $$ = $1 && $2; }
		| day_of_week_ok time_of_day_ok
			{ $$ = $1 && $2; }
		| day_of_year_ok day_of_week_ok
			{ $$ = $1 && $2; }
		| day_of_week_ok day_of_year_ok
			{ $$ = $1 && $2; }
		| day_of_year_ok
			{ $$ = $1; }
		| day_of_week_ok
			{ $$ = $1; }
		| time_of_day_ok
			{ $$ = $1; }
		| ANY
			{ $$ = 1; }
		| NONE
			{ $$ = 0; }
		| /* EMPTY */
			{ $$ = 1; }
		;

day_of_week_ok	: DAY_OF_WEEK
			{ $$ = intime(&$1.lo, &$1.hi); }
		| DAY_OF_WEEK DASH DAY_OF_WEEK
			{ $$ = intime(&$1.lo, &$3.hi); }
		| day_of_week_ok COMMA day_of_week_ok
			{ $$ = $1 || $3; }
		;

time_of_day_ok	: time_of_day
			{ $$ = intime(&$1.lo, &$1.hi); }
		| time_of_day DASH time_of_day
			{ $$ = intime(&$1.lo, &$3.hi); }
		| time_of_day_ok COMMA time_of_day_ok
			{ $$ = $1 || $3; }
		;

day_of_year_ok	: day_of_year
			{ $$ = intime(&$1.lo, &$1.hi); }
		| day_of_year DASH day_of_year
			{ $$ = intime(&$1.lo, &$3.hi); }
		| day_of_year_ok COMMA day_of_year_ok
			{ $$ = $1 || $3; }
		;

time_of_day	: TIME_OF_DAY
			{ $$ = $1; }
		| NUMBER COLON NUMBER COLON NUMBER MERIDIAN
			{
			  /*
			   * set up the hour, minute, and second
			   * and verify that the time is legal
			   */
			  $$ = nulltime;
			  $$.lo.tm_hour = $1 + $6.lo.tm_hour;
			  $$.hi.tm_hour = $1 + $6.hi.tm_hour;
			  $$.lo.tm_min = $$.hi.tm_min = $3;
			  $$.lo.tm_sec = $$.hi.tm_sec = $5;
			  if (badtime(&$$.hi) || badtime(&$$.lo))
				YYERROR;
			}
		| NUMBER COLON NUMBER COLON NUMBER
			{
			  /*
			   * set up the hour, minute, and second
			   * and verify that the time is legal
			   */
			  $$ = nulltime;
			  $$.lo.tm_hour = $$.hi.tm_hour = $1;
			  $$.lo.tm_min = $$.hi.tm_min = $3;
			  $$.lo.tm_sec = $$.hi.tm_sec = $5;
			  if (badtime(&$$.hi) || badtime(&$$.lo))
				YYERROR;
			}
		| NUMBER COLON NUMBER MERIDIAN
			{
			  /*
			   * set up the hour and minute and verify
			   * that the time is legal
			   */
			  $$ = nulltime;
			  $$.lo.tm_hour = $1 + $4.lo.tm_hour;
			  $$.hi.tm_hour = $1 + $4.hi.tm_hour;
			  $$.lo.tm_min = $$.hi.tm_min = $3;
			  if (badtime(&$$.hi) || badtime(&$$.lo))
				YYERROR;
			}
		| NUMBER COLON NUMBER
			{
			  /*
			   * set up the hour and minute and verify
			   * that the time is legal
			   */
			  $$ = nulltime;
			  $$.lo.tm_hour = $$.hi.tm_hour = $1;
			  $$.lo.tm_min = $$.hi.tm_min = $3;
			  if (badtime(&$$.hi) || badtime(&$$.lo))
				YYERROR;
			}
		| NUMBER MERIDIAN
			{
			  /*
			   * set up the hour and verify that the time is legal
			   */
			  $$ = nulltime;
			  $$.lo.tm_hour = $1 + $2.lo.tm_hour;
			  $$.hi.tm_hour = $1 + $2.hi.tm_hour;
			  if (badtime(&$$.hi) || badtime(&$$.lo))
				YYERROR;
			}
		| NUMBER 
			{
			  /*
			   * set up the hour and verify that the time is legal
			   */
			  $$ = nulltime;
			  $$.hi.tm_hour = $$.lo.tm_hour = $1;
			  if (badtime(&$$.hi) || badtime(&$$.lo))
				YYERROR;
			}
		;

day_of_year	: MONTH NUMBER COMMA NUMBER
			{
			  int yr;	/* normalized year */

			  /*
			   * get the month and set the day
			   * of the month and the year
			   */
			  if (badyear($4, &yr))
				YYERROR;
			  $$.lo.tm_mday = $$.hi.tm_mday = $2;
			  $$.lo.tm_year = $$.hi.tm_year = yr;
			}
		| MONTH NUMBER
			{
			  /*
			   * get the month and set the day
			   * of the month
			   */
			  $$ = $1;
			  $$.lo.tm_mday = $$.hi.tm_mday = $2;
			}
		| MONTH COMMA NUMBER
			{
			  int yr;	/* normalized year */

			  /*
			   * get the month and set the year
			   */
			  if (badyear($3, &yr))
				YYERROR;
			  $$ = $1;
			  $$.lo.tm_year = $$.hi.tm_year = yr;
			}
		| MONTH
			{ $$ = $1; }
		| NUMBER SLASH NUMBER SLASH NUMBER
			{
			  int yr;	/* normalized year */

			  /*
			   * get the month and set the day
			   * of the month and the year
			   */
			  if (badmonth($1, &$$) || badyear($5, &yr))
				YYERROR;
			  $$.lo.tm_mday = $$.hi.tm_mday = $3;
			  $$.lo.tm_year = $$.hi.tm_year = yr;
			}
		| NUMBER SLASH NUMBER
			{
			  /*
			   * get the month and set the day of the month
			   */
			  if (badmonth($1, &$$))
				YYERROR;
			  $$.lo.tm_mday = $$.hi.tm_mday = $3;
			}
		;

%%

/*
 * this is the lexer -- it's pretty dumb
 */
yylex()
{
	register int i;		/* used for table lookups if need be */

	/*
	 * this is hit at the end of string
	 * we need to do it this way because the '\0' (EOL)
	 * token must be returned, so we have to return
	 * another "end of input" token -- in other words,
	 * the end of input character is NOT the same as
	 * the end of string (EOL) character
	 */
	if (ateol){
		ateol = 0;
		return(-1);	/* YACC's end of file character */
	}

	/*
	 * eat leading white spaces
	 */
	while(*lptr && isspace(*lptr))
		lptr++;

	/*
	 * hit end of string character
	 * indicate there's nothing more with ateol
	 * (so next time we return YACC's end of file)
	 * and return the EOL token
	 */
	if (*lptr == '\0'){
		ateol = 1;
		return(EOL);
	}
	/*
	 * word -- look in reserved word area
	 */
	if (isalpha(*lptr)){
		if ((i = lookup(&lptr)) != LK_NOSUCH && i != LK_AMBIG){
			yylval.vval.lo = timelist[i].lo;
			yylval.vval.hi = timelist[i].hi;
			return(wordlist[i].rettype);
		}
	}

	/*
	 * number -- just return it
	 */
	if (isdigit(*lptr)){
		for(i = 0; isdigit(*lptr); lptr++)
			i = i * 10 + *lptr - '0';
		yylval.ival = i;
		return(NUMBER);
	}

	/*
	 * something else -- analyze it
	 */
	switch(*lptr++){
	case '(':		/* begin grouping */
		return(LPAR);
	case ')':		/* end grouping */
		return(RPAR);
	case '!':		/* negation */
		return(NOT);
	case '|':		/* conjunction */
		return(OR);
	case '&':		/* disjunction */
		return(AND);
	case '-':		/* interval indicator */
		return(DASH);
	case ',':		/* list separator, etc. */
		return(COMMA);
	case ':':		/* time separator */
		return(COLON);
	case '/':		/* date separator */
		return(SLASH);
	}

	/*
	 * unknown
	 */
	return(UNK);
}

/*==================== S U P P O R T   F U N C T I O N S =================*/

/*
 * special word lookup
 * the argument pointer is advanced to beyond the end of the match
 * note case doesn't matter
 * also note it can be any prefix that's unique; so,
 * "T" is no good ("Tuesday" and "Thursday") but "Th" is fine
 */
int lookup(s)
char **s;			/* string to be looked up */
{
	register int i;		/* counter in a for loop */
	register int uniq = LK_NOSUCH;	/* number of matches in table */
	register char *suniq = NULL;	/* number of matches in table */
	char *s1, *p;		/* used to do the comparisons */

	/*
	 * since we need to save s, we use s1 for comparisons
	 */
	s1 = *s;

	/*
	 * walk the list of the word
	 */
	for(i = 0; wordlist[i].string != NULL; i++){
		/*
		 * compare the current table entry and the string
		 * map the input case to lower case
		 * stop at first mismatch
		 */
		p = wordlist[i].string;
		while(*s1 && *p && lowcase(*s1) == *p){
			s1++;
			p++;
		}
		/*
		 * if no match, just loop
		 * if a match, next input char must NOT be a letter
		 * or there's no match
		 */
		if (s1 == *s)
			continue;
		if (*p == '\0' || *s1 == '\0' || !isalpha(*s1)){
			/*
			 * match -- if unmatched so far, remember it
			 * otherwise mark this as ambiguous and
			 * return
			 */
			if (uniq == LK_NOSUCH){
				uniq = i;
				suniq = s1;
			}
			else
				return(LK_AMBIG);
		}
		s1 = *s;
	}

	/*
	 * got something unambiguous
	 */
	if (uniq != LK_NOSUCH){
		*s = suniq;
		return(uniq);
	}

	/*
	 * nothing
	 */
	return(LK_NOSUCH);
}

/*
 * looks up something based on rettype and refnum
 * useful when you know the number of a month but not the name (for example)
 * this pair is always unique so we don't worry about ambiguities
 */
int nlookup(num, what)
int num;			/* number of rettype */
int what;			/* the rettype */
{
	register int i;		/* counter in a for loop */

	/*
	 * walk the lookup table
	 */
	for(i = 0; wordlist[i].string != NULL; i++)
		if (wordlist[i].rettype == what && wordlist[i].refnum == num){
			/*
			 * found it
			 */
			return(i);
		}
	/*
	 * nothing there ...
	 */
	return(LK_NOSUCH);
}

/*
 * map the number to a month
 * return 1 if mapping fails, 0 otherwise
 */
badmonth(n, ts)
int n;			/* number of month */
TIME *ts;		/* pointer to return structure */
{
	register int i;		/* holds value of month lookup */
	char buf[BUFSIZ];	/* for any error messages */

	/*
	 * see if the month is there
	 * if not, give error message
	 */
	if (n < 1 || n > 12 || (i = nlookup(n, MONTH)) == LK_NOSUCH){
		(void) sprintf(buf, "bad month %d", n);
		yyerror(buf);
		return(1);
	}

	/*
	 * return the associated structure
	 */
	*ts = timelist[i];
	return(0);
}
	
/*
 * normalize the year number
 * return 1 if the number is not valid
 * note any non-negative years are errors!
 */
badyear(year, norm)
int year;		/* year number */
int *norm;		/* where the normalized year number goes */
{
	char buf[BUFSIZ];	/* for any error messages */

	/*
	 * if year is negative,
	 * it's an error
	 */
	if (year < 0){
		(void) sprintf(buf, "bad year %d", year);
		yyerror(buf);
		return(1);
	}

	/*
	 * anything else is okay -- if over 99, normalize
	 */
	if (year < 100){
		*norm = year;
		return(0);
	}
	*norm = year - 1900;
	return(0);
}

/*
 * look for bad hour, minute, second
 */
badtime(ts)
struct tm *ts;			/* time to be checked */
{
	char buf[BUFSIZ];	/* for any error messages */

	/*
	 * check hours
	 */
	if (ts->tm_hour != -1 && ts->tm_hour > 23){
		(void) sprintf(buf, "bad hour %d", ts->tm_hour);
		yyerror(buf);
		return(1);
	}

	/*
	 * check minutes
	 */
	if (ts->tm_min != -1 && ts->tm_min > 59){
		(void) sprintf(buf, "bad minute %d", ts->tm_min);
		yyerror(buf);
		return(1);
	}

	/*
	 * check seconds
	 */
	if (ts->tm_sec != -1 && ts->tm_sec > 59){
		(void) sprintf(buf, "bad second %d", ts->tm_sec);
		yyerror(buf);
		return(1);
	}

	/*
	 * it's a good time
	 */
	return(0);
}

/*
 * time comparison function; return
 *		-1	if a < b
 *		 0	if a = b
 *		+1	if a > b
 * just like strcmp
 *   comparison is on year, then month, then day of the month,
 * then hour, then minute, then second -- ignoring any fields with
 * -1 in either a or b
 */
int timecmp(a, b)
struct tm *a, *b;		/* times to be compared */
{
	/*
	 * compare days of the week
	 */
	if (a->tm_wday != -1 && b->tm_wday != -1){
		if (a->tm_wday > b->tm_wday)		return(1);
		else if (a->tm_wday < b->tm_wday)	return(-1);
	}
	/*
	 * compare years
	 */
	if (a->tm_year != -1 && b->tm_year != -1){
		if (a->tm_year > b->tm_year)		return(1);
		else if (a->tm_year < b->tm_year)	return(-1);
	}
	/*
	 * compare months
	 */
	if (a->tm_mon != -1 && b->tm_mon != -1){
		if (a->tm_mon > b->tm_mon)		return(1);
		else if (a->tm_mon < b->tm_mon)		return(-1);
	}
	/*
	 * compare days of the month
	 */
	if (a->tm_mday != -1 && b->tm_mday != -1){
		if (a->tm_mday > b->tm_mday)		return(1);
		else if (a->tm_mday < b->tm_mday)	return(-1);
	}
	/*
	 * compare hours
	 */
	if (a->tm_hour != -1 && b->tm_hour != -1){
		if (a->tm_hour > b->tm_hour)		return(1);
		else if (a->tm_hour < b->tm_hour)	return(-1);
	}
	/*
	 * compare minutes
	 */
	if (a->tm_min != -1 && b->tm_min != -1){
		if (a->tm_min > b->tm_min)		return(1);
		else if (a->tm_min < b->tm_min)		return(-1);
	}
	/*
	 * compare seconds
	 */
	if (a->tm_sec != -1 && b->tm_sec != -1){
		if (a->tm_sec > b->tm_sec)		return(1);
		else if (a->tm_sec < b->tm_sec)		return(-1);
	}
	/*
	 * equality
	 */
	return(0);
}

/*= I N I T I A L I Z A T I O N   A N D   C H E C K I N G   R O U T I N E S =*/
/*
 * initialize the current time
 */
inittime()
{
	/*
	 * get the current time
	 */
	clocktm = time((long *)0);
	cur = *localtime(&clocktm);
#ifdef DEBUG
	/*
	 * print it if need be
	 */
	prtime(&cur, "CURRENT TIME");
#endif
}

/*
 * this routine sets up the string for parsing, calls the parser,
 * and handles the result
 */
isintime(t)
char *t;		/* buffer for time */
{
	/*
	 * get the current time
	 */
	inittime();

	/*
	 * set up the pointer to the input
	 * for yylex, the lexical analyzer
	 */
	lptr = &t[strlen(t)-1];
	if (*lptr == '\n')
		*lptr = '\0';
	lptr = t;

	/*
	 * parse the date and process the result
	 */
	if (yyparse()){
#ifdef DEBUG
		printf("illegal time description\n");
#endif
		return(0);
	}
#ifdef DEBUG
	printf("time %s this description\n",
				timing ? "matches" : "does not match");
#endif
	return(timing);
}

/*
 * error in parse
 * we just log it; since we do NOT do error recovery,
 * the attempt is rejected
 */
yyerror(s)
char *s;		/* error message (supplied by YACC) */
{
	char buf[BUFSIZ];	/* error message */
	register char *p;	/* used to format error message */
	extern int linect;	/* line number */

	/*
	 * shoot the rest of the line
	 */
	if ((p = index(s, '\t')) != NULL)
		*p = '\0';
	/*
	 * dump the error message
	 */
	(void) sprintf(buf, "(%2d): %s -- bad time (at \"%s\")\n",
							linect, s, --lptr);
	err(L_INFO|F_COND, buf);
}

/*===================== M A I N  C A L L I N G  R O U T I N E S ==============*/
/*
 * two sets of main routines
 * if "DEBUG" is defined, you get a main routine that reads in one date,
 * parses it, and prints the result
 * if "DEBUG" is not defined, you get a routine that returns 1 if the current
 * time is within the time range of the argument string, 0 if not
 *
 * "DEBUG" san be set to two levels; "1" gives you just the result (1 or 0),
 * but you can use the print function "prtime()" to print out key times in
 * the parse.  "2" gives you complete debugging info from YACC as well
 */

#ifdef DEBUG

/*
 * set the flag YYDEBUG
 */
#if DEBUG > 1
#define YYDEBUG		/* flag so YACC will generate debugging info */
extern int yydebug;	/* constant to tell YACC to generate debugging info */
#endif

int linect = 0;		/* number of expression being tested */

/*
 * main routine -- set current time, read the date, parse it,
 * and print the result
 */
main()
{
	char buf[BUFSIZ];		/* input buffer */

#if DEBUG > 1
	yydebug = 1;		/* YACC is to give full debugging output */
#endif

	/*
	 * get the input; if EOF, quit
	 */
	while(fgets(buf, BUFSIZ, stdin) != NULL){
		/*
		 * new expression
		 */
		linect++;

		/*
		 * clear the end of line flag
		 */
		ateol = 0;

		/*
		 * clobber any trailing newline
		 */
		lptr = &buf[strlen(buf)-1];
		if (*lptr == '\n')
			*lptr = '\0';
	
		/*
		 * print the buffer
		 */
		printf("buf is <%s>\n", buf);


		/*
		 * set up the pointer to the input
		 * for yylex, the lexical analyzer
		 */
		lptr = buf;

		/*
		 * parse the date and process the result
		 */
		(void) isintime(buf);
	}

	/*
	 * no problem
	 */
	exit(0);
}

/*
 * print the given time, preceeded by the given string,
 * in a readable format
 */
prtime(s, lb)
struct tm *s;		/* time */
char *lb;		/* label string */
{
	printf("%s %d/%d/%d %d:%02d:%02d (%d)\n", lb,
		s->tm_mon, s->tm_mday, s->tm_year,
		s->tm_hour, s->tm_min, s->tm_sec,
		s->tm_wday);
}

/*
 * error in parse
 * this mimics a function in lsu
 */
err(f, s)
unsigned int f;		/* flag */
char *s;		/* error message (supplied by YACC) */
{
	fprintf(stderr, "%s\n", s);
}

#endif

