%{
/*
 * this file contains the parser fot the terminal
 * the function "isintty(t) returns 1 id the current tty
 * matches the description given by the string t
 * grammar:
 *	ttyname <or> speed
 * where:
 *	ttyname == filename				(ie, "/dev/ttyi9")
 *	speed   == ( '=' | '@' ) number			(ie, "@9600")
 *		   '<=' number				(ie, "<=9600")
 *		   '>=' number				(ie, ">=9600")
 *		   ( '!=' | '<>' | '><' ) number	(ie, "!=9600")
 *		   '<' number				(ie, ">9600")
 *		   '>' number				(ie, ">9600")
 *	
 * if you want to change the syntax, you can debug the new grammar by defining
 * "DEBUG" -- this allows the file to be compiled as a separate program, and
 * will display the result.
 *
 * Author:
 *	Matt Bishop
 *	Research Institute for Advanced Computer Science
 *	NASA Ames Research Center
 *	Moffett Field, CA  94035
 *
 *	mab@riacs.arpa			ARPA
 *	...!decvax!decwrl!riacs!mab	UUCP
 *	...!ihnp4!ames!riacs!mab	UUCP
 *
 * Copyright (c) 1986.
 */
#include <ctype.h>
#include <stdio.h>

#ifdef DEBUG
#	define	F_COND	0	/* dummy value; not used here */
#	define	L_INFO	0	/* dummy value; not used here */
#	include "sysdep.h"
#else
#	include "lsu.h"
#endif

/*
 * useful macros
 */
#define	DEFTTY	"/dev/null"	/* if you can't get a tty description use this */
				/* if c is upper case, make it lower case */
#define	lowcase(c)	(isupper((c))?tolower((c)):(c))
				/*
				 * test for equality in name of device
				 * for input, output, error, and all
				 * devices
				 */
#define	inname(val)	checkname(&(thistty.input), val)
#define	outname(val)	checkname(&(thistty.output), val)
#define	errname(val)	checkname(&(thistty.errput), val)
#define	allname(val)	(inname(val) && outname(val) && errname(val))
				/*
				 * test for pattern match in name of device
				 * for input, output, error, and all
				 * devices
				 */
#define	inpat(val)	checkpat(&(thistty.input), val)
#define	outpat(val)	checkpat(&(thistty.output), val)
#define	errpat(val)	checkpat(&(thistty.errput), val)
#define	allpat(val)	(inpat(val) && outpat(val) && errpat(val))
				/*
				 * test for a relationship to a speed
				 * for input, output, error, and all
				 * devices
				 */
#define	inspeed(val)	checkspeed(thistty.input.speed, val)
#define	outspeed(val)	checkspeed(thistty.output.speed, val)
#define errspeed(val)	checkspeed(thistty.errput.speed, val)
#define	allspeed(val)	(inspeed(val) && outspeed(val) && errspeed(val))
				/*
				 * these are the relations for SPEED
				 */
#define	GR	1		/* (tty speed) > (number) */
#define	LT	2		/* (tty speed) < (number) */
#define	EQ	3		/* (tty speed) = (number) */
#define	GE	4		/* (tty speed) >= (number) */
#define	LE	5		/* (tty speed) <= (number) */
#define	NE	6		/* (tty speed) != (number) */


/*
 * the terminal description information is a bit bulky -- basically,
 * names of the devices and speeds are obtained for each of the terminals
 * connected to stdin, stdout, and stderr; all this information is saved
 * in a TTY structure
 */
struct ttydesc {
	char *path;		/* full device name of tty */
	char *devname;		/* common name of this device */
	long speed;		/* speed */
};
typedef struct ttyinfo {
	struct ttydesc input;	/* input tty description */
	struct ttydesc output;	/* output tty description */
	struct ttydesc errput;	/* error tty description */
} TTY;

/*
 * speeds in the language are represented by a relation and a number
 * anything satisfying the relation with the number is acceptable;
 * relations are the usual =, <, >, !=, <=, >=, and the number goes
 * on the right, so (i.e.) a speed of less than 9600 baud is saved as
 * { 9600, < }
 */
typedef struct speedinfo {
	long rate;		/* requisite speed */
	int state;		/* relation (EQ, NE, ...) */
} SPEED;

%}

/*
 * for reasons due entirely to the way YACC processes input,
 * this has to go here since SPEED is used as a type in the
 * header file to declare yylval
 */
%union {
	char *fval;		/* as a character string */
	int ival;		/* as an integer */
	SPEED sval;		/* as a speed */
}

/*
 * tokens returned by the lexical analyzer yylex()
 */
%token <ival> AND		/* integer: ampersand */
%token <ival> ANY		/* integer: matches any tty */
%token <ival> EOL		/* integer: no more input */
%token <ival> ERRPUT		/* integer: look only at tty on stderr */
%token <ival> INPUT		/* integer: look only at tty on stdin */
%token <ival> LPAR		/* integer: left parenthesis */
%token <fval> NAME		/* char *:  name of tty device */
%token <ival> NONE		/* integer: matches no tty */
%token <ival> NOT		/* integer: exclamation point */
%token <ival> OR		/* integer: vertical bar */
%token <ival> OUTPUT		/* integer: look only at tty on stdout */
%token <fval> PATTERN		/* char *:  pattern for tty device */
%token <sval> RATE		/* SPEED:   speed of tty device */
%token <ival> RPAR		/* integer: right parenthesis */

/*
 * productions analyzed by the parser
 */
%type <ival> stat		/* integer: expr with an EOL token tacked on */
%type <ival> expr		/* integer: 1 if tty matches description so far */
%type <ival> ttyname		/* integer: 1 if tty name matches device name */
%type <ival> ttyspeed		/* integer: 2 if tty speed matches device speed */

/*
 * expression operators
 * these are arranged so NOT has highest precedence,
 * followed by OR and AND (which must be on the same line since
 * they are of equal precedence)
 */
%left OR AND
%right NOT

%{

/*
 * table of speeds
 * UNIX systems store speeds as indices into a table;
 * the elements stored in the table are the speed in baud rate;
 * since this varies from UNIX to UNIX, it is defined in "sysdep.h"
 * (fractional baud rates get rounded to the nearest long)
 */
long speedtab[] = { SPEEDS };	/* table of tty speeds */
#define	SZSPEEDTAB	(sizeof(speedtab)/sizeof(long))	/* number of elements */

/*
 * unless escaped, any of the following characters terminate a file name
 */
char *eofile = ",=><!()|&+-*";	/* file name terminators */

/*
 * variables
 */
static TTY thistty;		/* information about this tty */
static int ttyval;		/* 1 if tty is okay, 0 if not */
static int attab = 0;		/* 1 when you hit the end of string */
static char *lptr;		/* used to walk input string */

/*
 * dunctions
 */
long getnum();			/* read a number from the file */
char *index();			/* find first occurrence of char in string */
char *rindex();			/* find last occurrence of char in string */
char *strsave();		/* used to save a string */
char *ttyname();		/* full path name of device */

%}

/*
 * start analysis at state stat
 */
%start stat

%%

stat		: expr EOL
			{ ttyval = $1; }
		;

expr		: LPAR expr RPAR
			{ $$ = $2; }
		| NOT expr
			{ $$ = !$2; }
		| expr OR expr
			{ $$ = $1 || $3; }
		| expr AND expr
			{ $$ = $1 && $3; }
		| ttyname
			{ $$ = $1; }
		| ttyspeed
			{ $$ = $1; }
		| ANY
			{ $$ = 1; }
		| NONE
			{ $$ = 0; }
		| /* EMPTY */
			{ $$ = 1; }
		;

ttyname		: NAME
			{ $$ = allname($1); free($1); }
		| INPUT NAME
			{ $$ = inname($2); free($2); }
		| OUTPUT NAME
			{ $$ = outname($2); free($2); }
		| ERRPUT NAME
			{ $$ = errname($2); free($2); }
		| PATTERN
			{ $$ = allpat($1); free($1); }
		| INPUT PATTERN
			{ $$ = inpat($2); free($2); }
		| OUTPUT PATTERN
			{ $$ = outpat($2); free($2); }
		| ERRPUT PATTERN
			{ $$ = errpat($2); free($2); }
		;

ttyspeed	: RATE
			{ if (($$ = allspeed(&$1)) == -1) YYERROR; }
		| INPUT RATE
			{ if (($$ = inspeed(&$2)) == -1) YYERROR; }
		| OUTPUT RATE
			{ if (($$ = outspeed(&$2)) == -1) YYERROR; }
		| ERRPUT RATE
			{ if (($$ = errspeed(&$2)) == -1) YYERROR; }
		;

%%

/*
 * this is the lexer -- it's not too smart
 */
yylex()
{
	register char c;		/* current character */
	register char *f;		/* used to load file name */
	char fname[BUFSIZ];		/* buffer for file name */

	/*
	 * this is hit at the end of the tty string
	 * we need to do it this way because the '\t' (EOL)
	 * token must be returned, so we have to return
	 * another "end of input" token -- in other words,
	 * the end of input character is NOT the same as
	 * the end of string (EOL) character
	 */
	if (attab){
		attab = 0;
		return(-1);		/* YACC's end of file character */
	}

	/*
	 * eat leading white spaces
	 */
	while(*lptr && *lptr == ' ')
		lptr++;
	
	/*
	 * hit end of string character
	 * indicate there's nothing more with attab
	 * (so next time we return YACC's end of file)
	 * and return the EOL token
	 */
	if (*lptr == '\t' || *lptr == '\n' || *lptr == '\0'){
		attab = 1;
		return(EOL);
	}

	/*
	 * now return the token found
	 */
	switch(c = *lptr++){
	case '(':		/* begin grouping */
		return(LPAR);
	case ')':		/* end grouping */
		return(RPAR);
	case '|':		/* conjunction */
	case ',':		/* disjunction */
		return(OR);
	case '&':		/* disjunction */
		return(AND);
	case '+':		/* input tty only */
		return(INPUT);
	case '-':		/* output tty only */
		return(OUTPUT);
	case '*':		/* error tty only */
		return(ERRPUT);
	case '=':		/* equal to a speed */
	case '@':
		yylval.sval.state = EQ;
		yylval.sval.rate = getnum();
		return(RATE);
	case '<':		/* less than/not equal to a speed */
		if (*lptr == '='){
			/*
			 * less than or equal to
			 */
			lptr++;
			yylval.sval.state = LE;
		}
		else if (*lptr == '>'){
			/*
			 * not equal to
			 */
			lptr++;
			yylval.sval.state = NE;
		}
		else{
			/*
			 * just less than
			 */
			yylval.sval.state = LT;
		}
		/*
		 * get the speed
		 */
		yylval.sval.rate = getnum();
		return(RATE);
	case '>':		/* greater than/not equal to a speed */
		if (*lptr == '='){
			/*
			 * greater than or equal to
			 */
			lptr++;
			yylval.sval.state = GE;
		}
		else if (*lptr == '<'){
			/*
			 * not equal to
			 */
			lptr++;
			yylval.sval.state = NE;
		}
		else{
			/*
			 * just less than
			 */
			yylval.sval.state = GR;
		}
		/*
		 * get the speed
		 */
		yylval.sval.rate = getnum();
		return(RATE);
	case '!':		/* not equal to a speed/negation */
		if (*lptr == '='){
			/*
			 * not equal to
			 */
			lptr++;
			yylval.sval.state = NE;
			/*
			 * get the speed
			 */
			yylval.sval.rate = getnum();
			return(RATE);
		}
		/*
		 * more general negation
		 */
		return(NOT);
	case '"':		/* pattern match */
		/*
		 * set up the pointer and put the character
		 * at the beginning of the array
		 */
		f = fname;
		/*
		 * go until unescaped end-of-file char or tab
		 */
		while(*lptr && *lptr != '\t' && *lptr != '"'){
			if (*lptr == '\\' && lptr[1])
				lptr++;
			*f++ = *lptr++;
		}
		*f = '\0';
		if (*lptr == '"')
			lptr++;
		/*
		 * save it somewhere (needed in case of lookahead)
		 */
		yylval.fval = strsave(fname);
		return(PATTERN);
	default:		/* anything else is a filename */
		/*
		 * set up the pointer and put the character
		 * at the beginning of the array
		 */
		f = fname;
		*f++ = c;
		/*
		 * go until unescaped end-of-file char or space
		 */
		while(*lptr && !isspace(*lptr) && index(eofile, *lptr) == NULL){
			if (*lptr == '\\' && lptr[1])
				lptr++;
			*f++ = *lptr++;
		}
		*f = '\0';
		/*
		 * see if it is "any"
		 */
		if (strlen(fname) == 3 && lowcase(fname[0]) == 'a' &&
		    lowcase(fname[1]) == 'n' && lowcase(fname[2]) == 'y')
			return(ANY);
		/*
		 * see if it is "none"
		 */
		if (strlen(fname) == 4 && lowcase(fname[0]) == 'n' &&
		    lowcase(fname[1]) == 'o' && lowcase(fname[2]) == 'n' &&
		    				     lowcase(fname[3]) == 'e')
			return(NONE);
		/*
		 * save it somewhere (needed in case of lookahead)
		 */
		yylval.fval = strsave(fname);
		return(NAME);
	}
	/* NOTREACHED */
}

/*==================== S U P P O R T   F U N C T I O N S ==================*/

/*
 * compare the name of a terminal device to the real device's name
 */
checkname(real, t)
struct ttydesc *real;		/* description of the current tty */
char *t;			/* filename of the device */
{
	/*
	 * see if it is a full path name (ie, any '/' in it?)
	 * compare appropriately
	 */
	if (index(t, '/') == NULL)
		return(strcmp(t, real->devname) == 0);
	return(strcmp(t, real->path) == 0);
}

/*
 * pattern match a terminal device to the real device's name
 */
checkpat(real, t)
struct ttydesc *real;		/* description of the current tty */
char *t;			/* filename pattern of the device */
{
	/*
	 * compile the pattern
	 */
	(void) smatch(t);

	/*
	 * compare appropriately
	 */
	return(match(real->devname) || match(real->path));
}

/*
 * compare the current speed to the real device's speed
 */
checkspeed(tst, this)
long tst;			/* the current speed */
SPEED *this;			/* the speed and expected relationship */
{
	/*
	 * check for a bogus speed
	 */
	if (tst == -1)
		return(-1);

	/*
	 * now compare
	 */
	switch(this->state){
	case EQ:	return(tst == this->rate);
	case LE:	return(tst >= this->rate);
	case GE:	return(tst <= this->rate);
	case NE:	return(tst != this->rate);
	case LT:	return(tst < this->rate);
	case GR:	return(tst > this->rate);
	}

	/*
	 * bad relationship -- fail!
	 */
	return(-1);
}

/*
 * read in a number
 */
long getnum()
{
	register long l;	/* used to gather number */

	/*
	 * be sure there's something there
	 */
	while(isspace(*lptr))
		lptr++;
	if (!isdigit(*lptr))
		return(-1);
	/*
	 * loop through the digits
	 */
	for(l = 0; isdigit(*lptr); lptr++){
		l *= 10;
		switch(*lptr){
		case '0':	l += 0;		break;
		case '1':	l += 1;		break;
		case '2':	l += 2;		break;
		case '3':	l += 3;		break;
		case '4':	l += 4;		break;
		case '5':	l += 5;		break;
		case '6':	l += 6;		break;
		case '7':	l += 7;		break;
		case '8':	l += 8;		break;
		case '9':	l += 9;		break;
		}
	}
	/*
	 * return the number
	 */
	return(l);
}


/*= I N I T I A L I Z A T I O N   A N D   C H E C K I N G   R O U T I N E S =*/

/*
 * initialize the terminal status function
 */
inittty()
{
	T_TTY ttybuf;			/* tty structure */
	char *p;			/* used to get name */
	int indx;			/* index into tty speed array */

	/*
	 * get the current input tty state
	 */
	if (ioctl(fileno(stdin), IO_GTTY, &ttybuf) < 0){
		thistty.input.speed = -1;
		thistty.input.path = strsave(DEFTTY);
	}
	else{
#ifndef LINUX
		if ((indx = TTYISPEED(ttybuf)) < 0 || indx > SZSPEEDTAB)
#endif
			indx = -1;
		thistty.input.speed = speedtab[indx];
		thistty.input.path = strsave(ttyname(fileno(stdin)));
	}
	if ((p = rindex(thistty.input.path, '/')) == NULL)
		thistty.input.devname = thistty.input.path;
	else
		thistty.input.devname = p + 1;

	/*
	 * get the current output tty state
	 */
	if (ioctl(fileno(stdout), IO_GTTY, &ttybuf) < 0){
		thistty.output.speed = -1;
		thistty.output.path = strsave(DEFTTY);
	}
	else{
#ifndef LINUX
		if ((indx = TTYOSPEED(ttybuf)) < 0 || indx > SZSPEEDTAB)
#endif
			indx = -1;
		thistty.output.speed = speedtab[indx];
		thistty.output.path = strsave(ttyname(fileno(stdout)));
	}
	if ((p = rindex(thistty.output.path, '/')) == NULL)
		thistty.output.devname = thistty.output.path;
	else
		thistty.output.devname = p + 1;

	/*
	 * get the current error tty state
	 */
	if (ioctl(fileno(stderr), IO_GTTY, &ttybuf) < 0){
		thistty.errput.speed = -1;
		thistty.errput.path = strsave(DEFTTY);
	}
	else{
#ifndef LINUX
		if ((indx = TTYOSPEED(ttybuf)) < 0 || indx > SZSPEEDTAB)
#endif
			indx = -1;
		thistty.errput.speed = speedtab[indx];
		thistty.errput.path = strsave(ttyname(fileno(stderr)));
	}
	if ((p = rindex(thistty.errput.path, '/')) == NULL)
		thistty.errput.devname = thistty.errput.path;
	else
		thistty.errput.devname = p + 1;

#ifdef DEBUG
	/*
	 * print it if need be
	 */
	prtty(&thistty, "CURRENT TTY");
#endif
}

/*
 * this routine sets up the string for parsing, calls the parser,
 * and handles the results
 */
isintty(t)
char *t;		/* buffer for string */
{
	/*
	 * get the current tty description
	 */
	inittty();

	/*
	 * set up the pointer to the input
	 * for yylex, the lexical analyzer
	 */
	lptr = &t[strlen(t)-1];
	if (*lptr == '\n')
		*lptr = '\0';
	lptr = t;

	/*
	 * parse the date and process the result
	 */
	if (yyparse()){
#ifdef DEBUG
		printf("illegal terminal description\n");
#endif
		return(0);
	}
#ifdef DEBUG
	printf("terminal %s this description\n",
				ttyval ? "matches" : "does not match");
#endif
	return(ttyval);
}

/*
 * error in parse
 * we just log it; since we do NOT do error recovery,
 * the attempt is rejected
 */
yyerror(s)
char *s;		/* error message (supplied by YACC) */
{
	char buf[BUFSIZ];	/* error message */
	register char *p;	/* used to format error message */
	extern int linect;	/* line number */

	/*
	 * truncate message at the end of field
	 */
	if ((p = index(lptr, '\t')) != NULL)
		*p = '\0';
	/*
	 * print the error message
	 */
	(void) sprintf(buf, "(%2d): %s -- bad tty (at \"%s\")\n",
							linect, s, --lptr);
	err(L_INFO|F_COND, buf);
}


/*===================== M A I N  C A L L I N G  R O U T I N E S ==============*/
/*
 * two sets of main routines
 * if "DEBUG" is defined, you get a main routine that reads in one date,
 * parses it, and prints the result
 * if "DEBUG" is not defined, you get a routine that returns 1 if the current
 * tty matches the description of the argument string, 0 if not
 *
 * "DEBUG" san be set to two levels; "1" gives you just the result (1 or 0),
 * but you can use the print function "prtty()" to print out key tty
 * descriptions in the parse.  "2" gives you complete debugging info from
 * YACC as well
 */

#ifdef DEBUG

/*
 * set the flag YYDEBUG
 */
#if DEBUG > 1
#define YYDEBUG		/* flag so YACC will generate debugging info */
extern int yydebug;	/* constant to tell YACC to generate debugging info */
#endif

int linect = 0;		/* number of expressions so far */

/*
 * main routine -- set current tty, read the description, parse it,
 * and print the result
 */
main()
{
	char buf[BUFSIZ];		/* input buffer */

#if DEBUG > 1
	yydebug = 1;		/* YACC is to give full debugging output */
#endif

	/*
	 * get the input; if EOF, quit
	 */
	while(fgets(buf, BUFSIZ, stdin) != NULL){
		/*
		 * another expression
		 */
		linect++;

		/*
		 * reset lexer state
		 */
		attab = 0;

		/*
		 * clobber any trailing newline
		 */
		lptr = &buf[strlen(buf)-1];
		if (*lptr == '\n')
			*lptr = '\0';
	
		/*
		 * print the buffer
		 */
		printf("buf is <%s>\n", buf);

		/*
		 * set up the pointer to the input
		 * for yylex, the lexical analyzer
		 */
		lptr = buf;

		/*
		 * parse the terminal and process the result
		 */
		(void) isintty(buf);
	}
}

/*
 * print the given tty information, preceeded by the given string,
 * in a readable format
 */
prtty(s, lb)
TTY *s;			/* tty */
char *lb;		/* label string */
{
	printf("%s (input)  %s (%s) %d baud\n",
		lb, s->input.path, s->input.devname, s->input.speed);
	printf("%s (output) %s (%s) %d baud\n",
		lb, s->output.path, s->output.devname, s->output.speed);
	printf("%s (error)  %s (%s) %d baud\n",
		lb, s->errput.path, s->errput.devname, s->errput.speed);
}

/*
 * error in parse
 * this mimics a function in lsu
 */
err(f, s)
unsigned int f;		/* flag */
char *s;		/* error message (supplied by YACC) */
{
	fprintf(stderr, "%s\n", s);
}

/*
 * allocate and copy a string into its own space
 * this mimics a function in lsu
 */
char *strsave(s)
char *s;		/* string to be saved */
{
	register char *p;		/* used to point to new space */
	char *malloc();			/* storage allocator */
	char *strcpy();			/* used to copy string */

	/*
	 * handle a NULL pointer properly
	 */
	if (s == NULL)
		s = "";
	/*
	 * allocate the space
	 */
	if ((p = malloc((unsigned)(strlen(s)+1))) == NULL)
		return(NULL);
	/*
	 * copy the string and return the new string
	 */
	(void) strcpy(p, s);
	return(s);
}

#endif

