/*
 * LSU -- local superuser
 *
 * Author:
 *	Matt Bishop
 *	Research Institute for Advanced Computer Science
 *	NASA Ames Research Center
 *	Moffett Field, CA  94035
 *
 *	mab@riacs.edu			ARPA
 *	...!decvax!decwrl!riacs!mab	UUCP
 *	...!ihnp4!ames!riacs!mab	UUCP
 *
 * Copyright (c) 1986.
 */
#include "lsu.h"

/*
 * setname -- get the tail of the path name; if legal, set progname to
 *	      the name in the "runas" array
 */
setname(path)
char *path;			/* full path name of program */
{
	register char *p, *q, *r;	/* used to find tail */
	register int i;			/* counter in a for loop */
	char buf[BUFSIZ];		/* buffer for error message */

	/*
	 * get the tail part of the name
	 */
	for(q = p = path; *p != '\0'; p++){
		if (*p == '/'){
			for(r = p; *p == '/'; p++);
			if (*p == '\0'){
				*r = '\0';
				break;
			}
			q = p;
		}
	}

	/*
	 * pick the element of runas[] that this program
	 * is being run as
	 */
	for(i = 0; runas[i].pname != NULL; i++)
		if (strcmp(q, runas[i].pname) == 0){
			progname = runas[i].pname;
			progis = &runas[i];
			return;
		}
	progname = q;
	progis = &runas[i];

	/*
	 * the program can only be called as SU or LSU
	 * die on anything else
	 */
	(void) sprintf(buf, "bad program name (%s)\n", path);
	err(L_INFO|F_FAIL, buf);
}

/*
 * isinlist -- see if arg 1 is in comma-separated list for arg2
 * if arg 1 is NULL, return the first element of the list in arg 1
 * if first element of list is "any", match
 */
char *isinlist(name, buf)
char *name;			/* name of person to check */
char *buf;			/* list to be checked */
{
	register char *b;	/* used to split comma-separated parts */
	register char *eob;	/* points to end of name list */

	/*
	 * find the end of the name list
	 */
	while(isspace(*buf))
		buf++;
	for(eob = buf; *eob && *eob != '\t'; eob++);
	if (*eob)
		*eob++ = '\0';
	/*
	 * if the list is "ANY", match
	 */
	if (lowcase(buf[0]) == 'a' &&
		 lowcase(buf[1]) == 'n' && lowcase(buf[2]) == 'y' &&
			(buf[3] == '\0' || buf[3] == '\t' || buf[3] == '\n'))
		return(eob);
	/*
	 * loop until you run out of list
	 */
	do{
		/*
		 * split the next name off from the comma-separated list
		 */
		for(b = buf; *b && *b != ','; b++);
		if (*b == ',')
			*b++ = '\0';
		/*
		 * compare -- see if this is it
		 */
		if (strcmp(name, buf) == 0)
			return(eob);
	}
	while(*(buf = b) != '\0');

	/*
	 * not found -- return nothing
	 */
	return(NULL);
}

/*
 * vfypwd -- check the password
 */
vfypwd(encpw)
char *encpw;		/* encrypted password */
{
	register int i;		/* used to blank out passwords */
	register char *p;	/* used to load passwords */

	/*
	 * demand a password if the stored password field is not null;
	 */
	if (curpw.pw_uid != 0 && encpw != NULL && *encpw != '\0'){
		/*
		 * we need a password 
		 */
		p = getpass("Password: ");
		if (success != F_FAIL && (p == NULL ||
	    		(strcmp(encpw, crypt(p, encpw)) != 0 &&
					strcmp(encpw, BOGUSPWD) != 0)))
				err(L_INFO|F_FAIL, "invalid password\n");
		/*
		 * blank out the password in core (just in case)
		 */
		for(i = 0; i < SZPASSWORD; i++)
			p[i] = '\001';
	}
}

/*
 * strsave -- saves a string in reserved space
 */
char *strsave(s)
char *s;			/* string to be copied */
{
	char *p;		/* pointer to new space */

	/*
	 * if there is no string, make an empty one
	 */
	if (s == NULL)
		s = "";

	/*
	 * allocate space for the string
	 */
	if ((p = malloc((unsigned) (strlen(s)+1))) == NULL)
		err(L_INFO|F_COND|F_SYS, "malloc");

	/*
	 * copy the string into the reserved memory
	 * and return the reserved memory space
	 */
	(void) strcpy(p, s);
	return(p);
}

/*
 * err -- print an error message (like perror)
 */
err(flag, what)
unsigned int flag;		/* error flags */
char *what;			/* what to print */
{
	char message[BUFSIZ];		/* message to print */

	/*
	 * if a system error, get the message
	 */
	if (bitset(flag, F_SYS)){
		if (errno < sys_nerr)
			(void) sprintf(message, "%s: %s\n",
						what, sys_errlist[errno]);
		else
			(void) sprintf(message, "%s: unknown error #%d\n",
						what, errno);
	}
	else
		(void) strcpy(message, what);

	/*
	 * if checking syntax, just report syntax errors
	 */
	if (progname == CSU){
		(void) fprintf(stderr, "%s: %s", progname, message);
		return;
	}

	/*
	 * log the error message
	 */
	logmessage(flag&L_MASK, message);

	/*
	 * mail a message
	 */
	if (bitset(flag, F_PERM))
		mailperm(message);
	if (bitset(flag, F_LOG))
		maillog(message);

	/*
	 * how to allow access
	 */
	if (bitset(flag, F_FAIL))
		success = F_FAIL;
	if (bitset(flag, F_COND) && (success != F_FAIL))
		success = F_COND;
}

/*
 * this mails a message to a list of users
 * who maintain the access file
 */
static char errperm[BUFSIZ] = "%s %s << xxEOFxx\n\
From: root\n\
To: %s\n\
Subject: \"su\" access file problems\n\
\n\
When %s tried to run \"%s\", it encountered a problem with the\n\
access file:\n\
	%s\
(The stamp for the log messages is [%05d].)  This must be fixed at once\n\
or the \"su\" programs will fail consistently.  As of now, only\n\
\"su root\" or \"nsu root\" can work.\n\
xxEOFxx\n";

mailperm(mesg)
char *mesg;			/* message */
{
	char buf[BUFSIZ];		/* buffer for mail message */

	(void) sprintf(buf, errperm, LSUMAIL, LSUMAINT, LSUMAINT,
					curpw.pw_name, progname, mesg, stamp);
	(void) system(buf);
}

/*
 * this mails a message to a list of users
 * who maintain the log file
 */
static char errlog[BUFSIZ] = "%s %s << xxEOFxx\n\
From: root\n\
To: %s\n\
Subject: \"su\" log file problems\n\
\n\
When %s tried to run \"%s\", it encountered a problem with the\n\
log file:\n\
	%s\
As a result, none of the \"su\" programs can log messages.  This\n\
must be fixed at once or the \"su\" programs will fail consistently.\n\
As of now, only \"su root\" or \"nsu root\" can work.\n\
xxEOFxx\n";

maillog(mesg)
char *mesg;			/* message */
{
	char buf[BUFSIZ];		/* buffer for mail message */

	(void) sprintf(buf, errlog, LSUMAIL, LSUMAINT, LSUMAINT,
					curpw.pw_name, progname, mesg);
	(void) system(buf);
}

